from PyQt5.QtWidgets import (
    QDialog, QWidget, QLabel, QSpinBox, 
    QComboBox, QDialogButtonBox, QGridLayout,
    QGroupBox, QCheckBox, QVBoxLayout,
    QLineEdit, QScrollArea, QSizePolicy
)
from PyQt5.QtCore import Qt

from plover_combo.combo_config import (
    ComboAlignment, ComboConfig, CONFIG_NAMES, CONFIG_ORDER, 
    CONFIG_TYPES, CONFIG_RANGES, ALIGNMENT_OPTIONS
)


FIELD_DATA_WIDTH = 250


class ConfigUI(QDialog):

    def __init__(self, temp_config: ComboConfig, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.temp_config = temp_config
        self.setup_window()

    def setup_window(self) -> None:
        self.scroll_widget = QWidget()
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(True)
        self.scroll_layout = QVBoxLayout()
        self.fields = dict()

        current_grid_row = 0
        current_grid_layout: QGridLayout = None
        current_groupbox: QGroupBox = None

        for config_name in CONFIG_ORDER:
            if config_name not in CONFIG_NAMES:
                if current_groupbox is not None:
                    current_groupbox.setLayout(current_grid_layout)
                    self.scroll_layout.addWidget(current_groupbox)
                
                current_grid_row = 0
                current_groupbox = QGroupBox()
                current_groupbox.setTitle(config_name)
                current_grid_layout = QGridLayout()
                continue
            
            field_label = QLabel()
            field_label.setText(CONFIG_NAMES[config_name])

            field_type = CONFIG_TYPES[config_name]
            field_data = None

            if field_type == bool:
                field_data = QCheckBox()
                field_data.setChecked(getattr(self.temp_config, config_name))

            elif field_type == int:
                low, high, step, suffix = CONFIG_RANGES[config_name]
                field_data = QSpinBox()
                field_data.setRange(low, high)
                field_data.setSingleStep(step)
                field_data.setSuffix(suffix)
                field_data.setValue(getattr(self.temp_config, config_name))

            elif field_type == str:
                field_data = QLineEdit()
                field_data.setText(getattr(self.temp_config, config_name))

            elif field_type == ComboAlignment:
                field_data = QComboBox()
                field_data.addItems(ALIGNMENT_OPTIONS)
                field_data.setCurrentIndex(
                    getattr(self.temp_config, config_name).value
                )

            if field_data is not None:
                field_data.setMinimumWidth(FIELD_DATA_WIDTH)
                current_grid_layout.addWidget(
                    field_label, current_grid_row, 0, 1, 1, Qt.AlignRight
                )
                current_grid_layout.addWidget(
                    field_data, current_grid_row, 1, 1, 1, Qt.AlignLeft
                )

                self.fields[config_name] = field_data
                current_grid_row += 1

        if current_groupbox is not None:
            current_groupbox.setLayout(current_grid_layout)
            self.scroll_layout.addWidget(current_groupbox)

        self.button_box = QDialogButtonBox(
            (
                QDialogButtonBox.Cancel | 
                QDialogButtonBox.Ok
            ),
            parent=self
        )
        self.button_box.rejected.connect(self.reject)
        self.button_box.accepted.connect(self.save_settings)

        
        self.scroll_widget.setLayout(self.scroll_layout)
        self.scroll_area.setWidget(self.scroll_widget)

        self.layout = QVBoxLayout()
        self.layout.addWidget(self.scroll_area)
        self.layout.addWidget(self.button_box)
        self.setLayout(self.layout)

    def save_settings(self) -> None:
        for config_name in CONFIG_NAMES.keys():
            field_type = CONFIG_TYPES[config_name]
            field_data = self.fields[config_name]
            field_value = None

            if field_type == bool:
                field_value = field_data.isChecked()
            elif field_type == int:
                field_value = field_data.value()
            elif field_type == str:
                field_value = field_data.text()
            elif field_type == ComboAlignment:
                field_value = ComboAlignment(field_data.currentIndex())

            if field_value is not None:
                setattr(self.temp_config, config_name, field_value)
        
        self.accept()



from PyQt5.QtWidgets import QLabel
from PyQt5.QtGui import QColor

# I am way too lazy to make this configurable.
BAR_ALPHA = 220
COLORS = {
    0:      (QColor(62, 167, 237),  QColor(106, 187, 241, BAR_ALPHA)),
    10:     (QColor(98, 221, 223),  QColor(136, 229, 231, BAR_ALPHA)),
    25:     (QColor(62, 221, 160),  QColor(99, 227, 178, BAR_ALPHA)),
    50:     (QColor(229, 189, 69),  QColor(235, 204, 112, BAR_ALPHA)),
    100:    (QColor(217, 114, 110), QColor(225, 145, 142, BAR_ALPHA)),
    250:    (QColor(227, 120, 166), QColor(234, 154, 188, BAR_ALPHA)),
    500:    (QColor(217, 61, 194),  QColor(225, 102, 206, BAR_ALPHA)),
    1000:   (QColor(147, 79, 219),  QColor(172, 120, 227, BAR_ALPHA)),
    2500:   (QColor(85, 81, 211),   QColor(128, 124, 222, BAR_ALPHA))
}


COLOR_NUMS = list(sorted(COLORS.keys()))


def round_to_checkpoint(num: int) -> int:
    if num <= 0:
        return 0

    prev = 0
    for color_num in COLOR_NUMS:
        if num < color_num:
            return prev
        
        prev = color_num
    
    return prev


def set_label_color(label: QLabel, color: QColor) -> None:
    values = "{r}, {g}, {b}, {a}".format(
        r = color.red(),
        g = color.green(),
        b = color.blue(),
        a = color.alpha()
    )

    label.setStyleSheet(f"color: rgba({values});")


import random
import sys

from plover.engine import StenoEngine
from plover.gui_qt.tool import Tool
from plover.gui_qt.utils import ToolBar
from plover.steno import Stroke

from PyQt5.QtWidgets import (
    QWidget, QPushButton, QGraphicsView, 
    QGraphicsScene, QApplication, QGraphicsTextItem,
    QGridLayout, QLabel, QSpacerItem, QSizePolicy,
    QGraphicsDropShadowEffect, QAction
)
from PyQt5.QtGui import (
    QMouseEvent, QFont, QKeyEvent, QPen, QBrush, 
    QFontDatabase, QColor, QKeySequence, QPainter,
    QPaintEvent
)
from PyQt5.QtCore import (
    Qt, QPoint, QVariantAnimation, QRectF, QSettings
)

from plover_combo.combo_colors import (
    COLORS, round_to_checkpoint, set_label_color
)
from plover_combo.combo_config import (
    CONFIG_ITEMS, CONFIG_TYPES, ComboAlignment, 
    ComboConfig
)
from plover_combo.config_ui import ConfigUI
from plover_combo.resources_rc import *


STYLESHEET = "border:0px; background:transparent;"


class ComboTool(Tool):
    TITLE = "Combo Counter"
    ICON = ":/combo/icon.svg"
    ROLE = "combo"

    def __init__(self, engine: StenoEngine) -> None:
        super().__init__(engine)
        self.setObjectName("combo")
        engine.signal_connect("stroked", self.on_stroke)

        QFontDatabase.addApplicationFont(":/combo/PloverRetro.ttf")

        self.drag_position = QPoint()
        self.counter = 0
        self.setting_highscore = False

        self.config = ComboConfig()
        self.restore_state()

        self.reload_config()
        self.setup_actions()
        self.setup_header()
        self.setup_counter()
        self.setup_cooldown_bar()
        self.setup_layout()

        self.finished.connect(self.save_state)

    def _restore_state(self, settings: QSettings) -> None:
        for field_name in CONFIG_ITEMS.keys():
            if settings.contains(field_name):
                setattr(
                    self.config, 
                    field_name, 
                    settings.value(field_name, type=CONFIG_TYPES[field_name])
                )

    def _save_state(self, settings: QSettings) -> None:
        for key, value in self.config.as_dict().items():
            settings.setValue(key, value)

    def paint_event(self, event: QPaintEvent) -> None:
        self.painter = QPainter(self)
        self.painter.setCompositionMode(QPainter.CompositionMode_Clear)
        self.painter.fillRect(self.rect(), Qt.transparent)
        self.painter.end()

    def on_settings(self) -> None:
        config_dialog = ConfigUI(self.config.copy(), self)
        if config_dialog.exec():
            self.config = config_dialog.temp_config
            self.reload_config()
    
    def on_stroke(self, stroke: Stroke) -> None:
        if self.config.reset_on_undo and stroke.is_correction:
            self.reset_counter()
            return
        
        self.increment_counter()

    def on_translate(self, undo: list, do: list, _) -> None:
        if self.config.reset_on_undo and len(undo) == 1:
            self.reset_counter()
            return
        
        self.increment_counter()

    def reload_config(self) -> None:
        self.zoom_scale = self.config.get_zoom_scale()
        self.counter_font = QFont(self.config.font_name, self.config.counter_font_size)
        self.title_font = QFont(self.config.font_name, self.config.title_font_size)
        self.subtitle_font = QFont(self.config.font_name, self.config.subtitle_font_size)
        self.main_color, self.sub_color = COLORS[0]

        if hasattr(self, "combo_header"):
            if self.config.dark_mode:
                set_label_color(self.combo_header, QColor(Qt.white))
            else:
                set_label_color(self.combo_header, QColor(Qt.black))

        if hasattr(self, "combo_header_shadow"):
            self.combo_header_shadow.setXOffset(self.config.shadow_x_offset)
            self.combo_header_shadow.setYOffset(self.config.shadow_y_offset)

    def setup_actions(self) -> None:
        self.close_action = QAction(self)
        self.close_action.setText("Close Widget")
        self.close_action.triggered.connect(self.accept)
        self.close_action.setShortcut(QKeySequence("Ctrl+X"))
        self.addAction(self.close_action)

        self.settings_action = QAction(self)
        self.settings_action.setText("Configure Widget")
        self.settings_action.triggered.connect(self.on_settings)
        self.settings_action.setShortcut(QKeySequence("Ctrl+S"))
        self.addAction(self.settings_action)

    def setup_header(self) -> None:
        self.combo_header = QLabel(self)
        self.combo_header.setText("Combo")
        self.combo_header.setFont(self.title_font)
        if self.config.dark_mode:
            set_label_color(self.combo_header, QColor(Qt.white))
        else:
            set_label_color(self.combo_header, QColor(Qt.black))

        self.combo_header_shadow = QGraphicsDropShadowEffect()
        self.combo_header_shadow.setBlurRadius(0)
        self.combo_header_shadow.setXOffset(self.config.shadow_x_offset)
        self.combo_header_shadow.setYOffset(self.config.shadow_y_offset)
        self.combo_header_shadow.setColor(self.main_color)
        self.combo_header.setGraphicsEffect(self.combo_header_shadow)
        
        self.highscore_header = QLabel(self)
        self.highscore_header.setText(f"HI {self.config.highscore}")
        self.highscore_header.setFont(self.subtitle_font)
        set_label_color(self.highscore_header, self.sub_color)

    def setup_counter(self) -> None:
        self.text_view = QGraphicsView(self)
        self.text_view.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.text_view.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.text_view.setStyleSheet(STYLESHEET)

        self.text_scene = QGraphicsScene(self)
        self.text_view.setScene(self.text_scene)
        self.counter_text = self.text_scene.addText(str(self.counter), self.counter_font)
        self.counter_text.setPos(self.config.horz_margin, self.config.top_margin)
        self.counter_text.setDefaultTextColor(self.main_color)

        self.text_view.mouseMoveEvent = self.view_mouse_move
        self.text_view.mousePressEvent = self.view_mouse_press
    
    def setup_cooldown_bar(self) -> None:
        self.cooldown_view = QGraphicsView(self)
        self.cooldown_view.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.cooldown_view.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.cooldown_view.setStyleSheet(STYLESHEET)

        self.cooldown_scene = QGraphicsScene(self)
        self.cooldown_view.setScene(self.cooldown_scene)
        self.cooldown_pen = QPen(Qt.NoPen)
        self.cooldown_brush = QBrush(self.sub_color)
        self.cooldown_bar = self.cooldown_scene.addRect(
            QRectF(),
            self.cooldown_pen,
            self.cooldown_brush
        )

        self.cooldown_view.mouseMoveEvent = self.view_mouse_move
        self.cooldown_view.mousePressEvent = self.view_mouse_press

    def setup_layout(self) -> None:
        self.setWindowFlags(self.windowFlags() | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setStyleSheet("QWidget#combo {background:transparent;}")

        self.layout = QGridLayout()
        self.layout.setContentsMargins(0, 10, 0, 10)
        self.layout.addWidget(self.combo_header, 0, 0, 1, 1, Qt.AlignCenter)
        self.layout.addWidget(self.highscore_header, 1, 0, 1, 1, Qt.AlignCenter)
        self.layout.addWidget(self.text_view, 2, 0, 2, 1)
        self.layout.addWidget(self.cooldown_view, 3, 0, 1, 1)
        self.setLayout(self.layout)

        self.mouseMoveEvent = self.view_mouse_move
        self.mousePressEvent = self.view_mouse_press
        self.paintEvent = self.paint_event

        self.animate()
        self.show()

    def view_mouse_move(self, event: QMouseEvent) -> None:
        if event.buttons() & Qt.LeftButton:
            self.move(event.globalPos() - self.drag_position)

    def view_mouse_press(self, event: QMouseEvent) -> None:
        if event.buttons() & Qt.LeftButton:
            self.drag_position = event.globalPos() - self.frameGeometry().topLeft()
    
    def increment_counter(self) -> None:
        self.counter += 1
        shaked = False

        if self.counter in COLORS:
            self.update_colors()
        
        self.animate()

        if self.counter in COLORS and self.counter > 0:
            shaked = True
            self.animate_shake()

        if self.counter > self.config.highscore:
            self.config.highscore = self.counter
            self.highscore_header.setText(f"HI {self.config.highscore}")
            if not (self.setting_highscore or shaked):
                self.animate_shake()

            self.setting_highscore = True
    
    def reset_counter(self) -> None:
        self.counter = 0
        if hasattr(self, "cooldown_animation") and self.cooldown_animation is not None:
            self.cooldown_animation.stop()
            self.cooldown_bar.setRect(QRectF())

        self.setting_highscore = False
        self.update_colors()
        self.animate()
    
    def update_colors(self) -> None:
        color_index = self.counter
        if self.counter not in COLORS:
            color_index = round_to_checkpoint(self.counter)
        
        self.main_color, self.sub_color = COLORS.get(color_index)
        self.combo_header_shadow.setColor(self.main_color)
        set_label_color(self.highscore_header, self.sub_color)
        self.counter_text.setDefaultTextColor(self.main_color)
        self.cooldown_brush.setColor(self.sub_color)
        self.cooldown_bar.setBrush(self.cooldown_brush)

    def animate(self) -> None:
        self.counter_text.setPlainText(str(self.counter))
        self.adjust_window()
        self.animate_counter()
        self.animate_cooldown()

    def adjust_window(self) -> None:
        text_bound = self.counter_text.boundingRect()
        self.text_width = text_bound.width()
        self.text_height = text_bound.height()

        self.width = int(self.text_width + self.config.horz_margin * 2)
        self.height = int(self.text_height + self.config.top_margin + self.config.bottom_margin)

        prev_geometry = self.frameGeometry()
        self.setFixedWidth(self.width)
        self.text_view.setFixedSize(self.width, self.height)
        self.text_scene.setSceneRect(0, 0, self.width, self.height)
        self.cooldown_view.setFixedSize(self.width, self.config.bar_width)
        self.cooldown_scene.setSceneRect(0, 0, self.width, self.config.bar_width)

        if self.config.alignment == ComboAlignment.CENTER:
            self.move(self.pos() - self.frameGeometry().center() + prev_geometry.center())
        elif self.config.alignment == ComboAlignment.RIGHT:
            self.move(self.pos() - self.frameGeometry().topRight() + prev_geometry.topRight())
        elif self.config.alignment == ComboAlignment.LEFT:
            self.move(self.pos() - self.frameGeometry().topLeft() + prev_geometry.topLeft())
        
        self.update()

    def animate_counter(self) -> None:
        self.counter_animation: QVariantAnimation
        if hasattr(self, "counter_animation") and self.counter_animation is not None:
            self.counter_animation.stop()
        
        self.counter_animation = QVariantAnimation(self.text_view)
        self.counter_animation.setStartValue(QRectF(
            self.width * (1.0 - self.zoom_scale) * 0.5, 
            self.height * (1.0 - self.zoom_scale), 
            self.width * self.zoom_scale, 
            self.height * self.zoom_scale
        ))
        self.counter_animation.setEndValue(QRectF(
            0, 0, self.width, self.height
        ))
        
        self.counter_animation.valueChanged.connect(
            lambda x: self.text_view.fitInView(x, Qt.KeepAspectRatio)
        )

        self.counter_animation.setDuration(self.config.counter_anim_duration)
        self.counter_animation.start()

    def animate_cooldown(self) -> None:
        if self.counter == 0:
            return

        self.cooldown_animation: QVariantAnimation
        if hasattr(self, "cooldown_animation") and self.cooldown_animation is not None:
            self.cooldown_animation.stop()
        
        self.cooldown_animation = QVariantAnimation(self.cooldown_view)
        self.cooldown_view.fitInView(QRectF(0, 0, self.width, self.config.bar_width))
        self.cooldown_animation.setStartValue(QRectF(
            self.config.horz_margin, 0, self.text_width, self.config.bar_width
        ))
        self.cooldown_animation.setEndValue(QRectF(
            self.width/2, 0, 0, self.config.bar_width
        ))
        self.cooldown_animation.valueChanged.connect(
            lambda x: self.cooldown_bar.setRect(x)
        )
        self.cooldown_animation.finished.connect(self.reset_counter)

        self.cooldown_animation.setDuration(self.config.cooldown_duration)
        self.cooldown_animation.start()
    
    def animate_shake(self) -> None:
        if not self.config.shake_enabled:
            return

        shake_animation = QVariantAnimation(self)

        animation_time = 0.0
        shake_interval = 1.0 / self.config.shake_count
        for count in range(self.config.shake_count):
            shake_animation.setKeyValueAt(
                animation_time, 
                QPoint(
                    self.pos().x() + random.randint(-1, 1) * self.config.shake_intensity,
                    self.pos().y() + random.randint(-1, 1) * self.config.shake_intensity,
                )
            )
            animation_time += shake_interval
        
        shake_animation.valueChanged.connect(
            lambda x: self.move(x)
        )
        shake_animation.setEndValue(QPoint(self.pos()))
        shake_animation.setDuration(self.config.shake_duration)
        shake_animation.start()


# if __name__ == "__main__":
#     app = QApplication(sys.argv)
#     window = ComboTool()

#     app.exec()



from enum import Enum
from typing import Callable


class ComboAlignment(Enum):
    LEFT = 0
    CENTER = 1
    RIGHT = 2


ALIGNMENT_OPTIONS = [
    "Align Left (Expand Right)",
    "Align Center",
    "Align Right (Expand Left)"
]


CONFIG_ITEMS = {
    "highscore": 0,
    "cooldown_duration": 2000,
    "reset_on_undo": True,
    "top_margin": 10,
    "bottom_margin": 10,
    "horz_margin": 30,
    "bar_width": 12,
    "zoom_scale_percent": 84,
    "alignment": ComboAlignment.CENTER,
    "font_name": "Plover Retro",
    "dark_mode": False,
    "title_font_size": 15,
    "subtitle_font_size": 12,
    "counter_font_size": 65,
    "shadow_x_offset": 2,
    "shadow_y_offset": 2,
    "counter_anim_duration": 180,
    "shake_enabled": True,
    "shake_duration": 250,
    "shake_count": 20,
    "shake_intensity": 3
}


CONFIG_TYPES = {k: type(v) for k, v in CONFIG_ITEMS.items()}


class ComboConfig:
    def __init__(self, values: dict = None):
        if values is None:
            values = dict()

        for key, default in CONFIG_ITEMS.items():
            if key in values:
                setattr(self, key, values[key])
            else:
                setattr(self, key, default)

    def copy(self) -> "ComboConfig":
        value_dict = {k: getattr(self, k) for k in CONFIG_ITEMS.keys()}
        return ComboConfig(value_dict)

    def get_zoom_scale(self) -> float:
        return self.zoom_scale_percent / 100
    
    def as_dict(self) -> dict:
        results = dict()
        for key in CONFIG_ITEMS.keys():
            results[key] = getattr(self, key)
        
        return results


CONFIG_NAMES = {
    "cooldown_duration": "Cooldown Duration",
    "reset_on_undo": "Reset Combo on Undo",
    "top_margin": "Counter Top Margin",
    "bottom_margin": "Counter Bottom Margin",
    "horz_margin": "Counter Horizontal Margin",
    "bar_width": "Cooldown Bar Thickness",
    "zoom_scale_percent": "Counter Normal to Enlarged Ratio",
    "alignment": "Widget Alignment",
    "font_name": "Font Name",
    "dark_mode": "Dark Mode (Title Color)",
    "title_font_size": "Title Font Size",
    "subtitle_font_size": "Highscore Font Size",
    "counter_font_size": "Counter Font Size",
    "shadow_x_offset": "Title Shadow Horizontal (X) Offset",
    "shadow_y_offset": "Title Shadow Vertical (Y) Offset",
    "counter_anim_duration": "Counter Animation Duration",
    "shake_enabled": "Enable Shake Animation",
    "shake_duration": "Shake Animation Duration",
    "shake_count": "Shake Speed",
    "shake_intensity": "Shake Intensity"
}


CONFIG_ORDER = [
    "Combo Settings",
    "reset_on_undo",
    "cooldown_duration",

    "Display Settings",
    "top_margin",
    "bottom_margin",
    "horz_margin",
    "bar_width",
    "zoom_scale_percent",
    "alignment",
    
    "Font Settings",
    "dark_mode",
    "font_name",
    "title_font_size",
    "subtitle_font_size",
    "counter_font_size",
    "shadow_x_offset",
    "shadow_y_offset",

    "Animation Settings",
    "shake_enabled",
    "shake_duration",
    "shake_count",
    "shake_intensity",
    "counter_anim_duration"
]


CONFIG_RANGES = {
    "cooldown_duration": (100, 30000, 100, "ms"),
    "top_margin": (-300, 300, 1, "px"),
    "bottom_margin": (-300, 300, 1, "px"),
    "horz_margin": (-300, 300, 1, "px"),
    "bar_width": (1, 300, 1, "px"),
    "zoom_scale_percent": (1, 100, 1, "%"),
    "title_font_size": (1, 300, 1, None),
    "subtitle_font_size": (1, 300, 1, None),
    "counter_font_size": (1, 300, 1, None),
    "shadow_x_offset": (-100, 100, 1, "px"),
    "shadow_y_offset": (-100, 100, 1, "px"),
    "counter_anim_duration": (10, 500, 10, "ms"),
    "shake_duration": (10, 500, 10, "ms"),
    "shake_count": (1, 100, 1, None),
    "shake_intensity": (1, 100, 1, "px")
}
